// Digital Engine Tachometer x86 Embedded Controller
//
// Author: Stephen Manley - smanley@nyx.net
// Date: November 1st 1999
//                                          
// Hardware:
//
// Intel 80188XL Embedded Processor
// Intel 8256A Multifunction I/O Controller (UART)
// ADC8020 High Speed Analog->Digital Converter
// Generic Output Latch
//
///////////////////////////////////////////////////////////////////////////

// System includes
#include<string.h>
#include<stdlib.h>

// Type definitions
typedef char *string;

// Device Addresses
#define PCS_BASE_ADDR	0x2000
#define TEST_LATCH  	   (PCS_BASE_ADDR + 0x0080)
#define UART 	      	(PCS_BASE_ADDR)
#define ADC					(PCS_BASE_ADDR + 0x0100)

// UART Miscellanea
#define UART_COMMAND1   (UART + 0x00)
#define UART_COMMAND2	(UART + 0x02)
#define UART_COMMAND3	(UART + 0x04)
#define UART_MODE 		(UART + 0x06)
#define UART_PORT1CTL	(UART + 0x08)
#define UART_SETINTS	   (UART + 0x0A)
#define UART_RESETINTS	(UART + 0x0C)
#define UART_RXBUFFER	(UART + 0x0E)
#define UART_TXBUFFER	(UART + 0x0E)
#define UART_PORT1		(UART + 0x10)
#define UART_STATUS		(UART + 0x1E)

#define CR					0x0D
#define LF					0x0A


// Timer Definitions
#define TCUCON 0xFF32

#define T0CNT 0xFF50
#define T0CON 0xFF56
#define T0CMPA 0xFF52
#define T0CMPB 0xFF54

#define T1INTNO 18
#define T1CNT  0xFF58
#define T1CMPA 0xFF5A
#define T1CMPB 0xFF5C
#define T1CON  0xFF5E

#define T2CNT  0xFF60
#define T2CMPA 0xFF62
#define T2CON  0xFF66

#define EOI    0xFF22
#define IMASK  0xFF28
#define PRIMASK 0xFF2A

// DMA Definitions
#define D0SRCH 0xFFC2
#define D0SRCL 0xFFC0
#define D0DSTH 0xFFC6
#define D0DSTL 0xFFC4
#define D0CON  0xFFCA
#define D0TC   0xFFC8

// Function definitions
void initUART(void);
void initPCS(void);
void sendChar_UART(char data);
void sendString_UART(char* string);
void displayADC(void);
void initTimer(void);
void timerStop(void);
void timerStart(void);
void timerBegin(void);
void displayProgramMenu(void);
void displayActive(void);
void displayMainMenu(void);
void testDMA(void);
int  checkUserAbort(void);
void calibrateTach(void);
void sampleDMA(void);

void outportw(unsigned int port, unsigned int value);
void outportb(unsigned int port, char value);

void far interrupt isrAlarm(void);

// Global Variables
char dummy[2000];
//int sampleRate = 0x50;
int sampleRate = 0x05;

int redlineRPM = 5000;
int shiftLightActive = 0;
int highestRPM = 0;
int numberCylinders = 4;
int numberTicks = 1;
int numberPulsesPerK = 35;
int pulseCount = 0;
int numberRPMPerK = 1000;

// Interrupts & System Init
///////////////////////////////////////////////////////////////////////////

//	Initialization of interrupt vector table
void initIVT(void)
{
	int timerAddr1 = T1INTNO*4 + 2;
   int timerAddr2 = T1INTNO*2;

	_asm
   {
   	push ds
      push ax
      xor ax,ax
      mov ds, ax
      cli
   	mov [timerAddr1], seg  isrAlarm
      mov [timerAddr2], offset isrAlarm
      sti
      pop ax
      pop ds
	}
}

// Timer interrupt
void far interrupt isrAlarm(void)
{
  //	int i;
   static j;

   if(j++<1000)
   	return;

	j=0;

   outportb(0x2080, 0x01); // in case we stuck


   // STOP TIMER 1
 	outportw(EOI,0x8012);  //send non-specific end-of-interrupt for type 18
}

void timerStart(void)
{
	outportw(T1CON, 0xE013); //Start Timer1, gen ints, cont.
}

void timerStop(void)
{
	outportw(T1CON, 0x4013);
}

// Set up timer interrupts on timer 1 for alarm processing
void timerBegin(void)
{
	timerStop();

   initIVT();
   outportw(TCUCON, 0x0004); 		// Enable timer ints, middle priority
   outportw(IMASK,  0x00FC);  	// Mask all interupt sources except TMR
   outportw(PRIMASK,0x0000); 		// Don't mask anything else (priority mask)

   timerStart();
}


void initTimer2(void)
{
   outportw(T2CNT,0); 					//Clear Timer Count
	outportw(T2CMPA, sampleRate);  	//Count A
	outportw(T2CON,0xC000);  			//Enable, dual mode, cont

// outportw(T0CNT,0); 					//Clear Timer Count
//	outportw(T0CMPA,7);  				//Count A
// outportw(T0CMPB,6);  				//Count B
//	outportw(T0CON,0xC003);  			//Enable, dual mode, cont
}

void timer2Start(void)
{
	outportw(T2CON,0xC001);  //Enable, cont
}

void timer2Stop(void)
{
	outportw(T2CON,0x4020);  //Disable, set max count
}

// Timing Functions
///////////////////////////////////////////////////////////////////////////

// Delay loop for LCD I/O
void delay(void)
{
	int i;
   for(i=0;i<100;i++)
   	_asm nop;
}

// Delays for an integral number of seconds, accurate to at least +-1 sec
void sleep(unsigned secs)
{
	long i;
	for(;secs;secs--)
   	for(i=0;i<10100;i++)
      	_asm nop;
}

/// Low Level IO
///////////////////////////////////////////////////////////////////////////

// Send byte to I/O space
void outportb(unsigned int port, char value)
{
	_asm {
      push ax
      push dx
   	mov dx, port
      mov al, value
      out dx, al
      pop dx
      pop ax
   }
   delay();
}

// Send word to I/O space
void outportw(unsigned int port, unsigned int value)
{
	_asm {
      push ax
      push dx
   	mov dx, port
      mov ax, value
      out dx, ax
      pop dx
      pop ax
   }
}

// Get byte from I/O space
char inportb(unsigned int port)
{
	char temp;
	_asm {
      push ax
      push dx
      mov dx, port
   	in al, dx
      mov temp, al
      pop dx
      pop ax
   }
   return temp;
}

// Get word from I/O space
unsigned int inportw(unsigned int port)
{
	unsigned int temp;
	_asm {
      push ax
      push dx
      mov dx, port
   	in ax, dx
      mov temp, ax
      pop dx
      pop ax
   }
   return temp;
}

// Initializes the UART
void initUART( void )
{
	// Command1 Register = 0100 0010 - x86 Mode, 1 Stop Bit, 7 bits/char
	// 										  Counter/Timer 2 mapped to priority level 1
   // 										  Port 1 P16 is a general purpose I/O port
   outportb(UART_COMMAND1, 0x42);  // 7 bit

   // Command2 Register = 0010 0100 - 9600 Baud Rate, 2.048 MHz CLK, Odd Parity
   // 										  No Parity Bit
	outportb(UART_COMMAND2, 0x24);

   // Command3 Register = 1100 0001 - Enable Serial Recieve
	//											  Reset status register, interrupts, etc
   outportb(UART_COMMAND3, 0xC1);

   // Mode Register = 0000 0110 - Do not Use Port 2
	outportb(UART_MODE, 0x06);

   // Port (Timing) 1 Control Register = 1111 1111
   // 											Sets p10 to p17 as outputs
   outportb(UART_PORT1CTL, 0xFF);
   //	outportb(UART_PORT1CTL, 0xF7);

   // NO Interrupts = 1111 1111
	outportb(UART_RESETINTS, 0xFF);
}


// Sends a char to the UART
void sendChar_UART(char data)
{
	// Wait for TBE (Transmit Buffer Empty) bit
	while(!(inportb(UART_STATUS) & 0x20));

	// Send data over the bus
	outportb(UART_TXBUFFER, data);
   return;
}

// Gets a char from the UART
char getChar_UART(void)
{
      //Wait for RBF (Receive Buffer Full) Bit
 		while(!(inportb(UART_STATUS) & 0x40));

      //Return contents of receive buffer)
      return (inportb(UART_TXBUFFER));
}


// Sends a string to the UART
void sendString_UART(char *string)
{
      while(*string != '\0')
      {
      	sendChar_UART(*string);
         string++;
      }
}

// Fills buffer with string information
void getString_UART(char* buffer)
{
	char a;
   int i = 0;
   int working = 1;

   while (working)
	{
   	a = getChar_UART();

      if ( (a == CR) || (a == LF))
			working = 0;
		else
	   	buffer[i] = a;

      i++;
   }

   buffer[i] = '\0';
}

void initDMA(void)
{
	int segment, off;
	outportw(D0SRCH, 0);				// Configure source address
	outportw(D0SRCL, ADC);

   outportw(D0DSTH, 0);				// Configure target address
 	_asm mov segment, seg dummy[0];
   _asm mov off, offset dummy[0];

   outportw(D0DSTL,(segment*16+off)&0x1FFF);

	outportw(D0TC,  2000);       // Number of bytes to transfer
   outportw(D0CON, 0xA270);	  // Sets byte transfer, not armed, from I/O to
//
   									  // memory, static source, incremented target

 //  outportw(D0CON, 0x
}

void armDMA(void)
{
	outportw(D0CON, 0xA276);
}

/*
void displayADC(void)
{
   int master = 0;
   int temp = 0;
   int working = 1;
   char a;

	sendString_UART("Detecting Pulses...\0");
   sendChar_UART(CR);
   sendChar_UART(LF);

   master = inportb(ADC);

   sleep (1);

	sendString_UART("Beginning sampling\0");
   sendChar_UART(CR);
   sendChar_UART(LF);

   while (working)
   {
   	temp = inportb(ADC);

   	if (0x04 < temp)
      {
   		sendString_UART("Voltage is high, pulse edge detected\0");
		   sendChar_UART(CR);
   		sendChar_UART(LF);

         // Wait
         while (0x04 < inportb(ADC));

	   	temp = inportb(ADC);
   	}
      else
      {
   		sendString_UART("Pulse level indeterminant or not connected\0");
  			sendChar_UART(CR);
	   	sendChar_UART(LF);

			// Wait
         while (0x04 > inportb(ADC));

	   	temp = inportb(ADC);
      }
   }
}
*/

// Assembles an output string for client software (pilot)
char* assembleOutputString(void)
{
 	return NULL;
}

// Displays a tachometer reading for self-client mode
char* displayTach(void)
{
	return NULL;
}

// Checks to see if there's a character in the buffer
int checkUserAbort(void)
{
    //Wait for RBF (Receive Buffer Full) Bit
    if ((inportb(UART_STATUS) & 0x40))
    	return 1;
    else
    	return 0;
}


// Grab the pulses and store them in pulseCount variable
void sampleDMA()
{
		int working = 1;
      int i = 0;


 	   initTimer2();
      initDMA();
      armDMA();
		timer2Start();

      while(inportw(D0TC)>0);

		i = 0;
      pulseCount = 0;

      working = 1;
 		while (working)
   	{
      	if (i >= 2000)
         	working = 0;

   		if (dummy[i] & 0x80)
      	{
         	// Wait
         	while (dummy[i] & 0x80)
            	i++;
   		}
      	else
      	{
         	// Wait
         	while (!(dummy[i] & 0x80))
            	i++;
  	    	}

         i++;
         pulseCount++;
	   }

      if (pulseCount <= 0)
      	pulseCount = 1;
}

void testDMA()
{
   int i;
   char temp;
   int working = 1;
   int measureCount = 0;
   char reference;
//   int pulseCount = 0;
   int pulseFrequency = 0;
   char buffer[20];
   char a;

   int errorInit = 0;
   int errorFrame = 0;

   int testing = 1;
   int val1, val2, set;

	// Initialize to unlikely value
 	for (i=0; i<2000; i++)
   	dummy[i] = 0x01;


   while (testing)
   {

   	sendString_UART("Acquring Voltage Reference for DMA Evaluation");
		sendChar_UART(CR);
  		sendChar_UART(LF);

   	reference =  inportb(ADC);
   	sendString_UART("Reference: ");

   	if (reference & 0x01)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x02)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x04)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x08)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x10)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x20)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

  		if (reference & 0x40)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

   	if (reference & 0x80)
			sendChar_UART('1');
 		else
   		sendChar_UART('0');

		sendChar_UART(CR);
  		sendChar_UART(LF);
		sendChar_UART(CR);
  		sendChar_UART(LF);

 		// Light up red LEDs
      outportb(0x2080, 0x03);

    	// Initialize DMA
      sendString_UART("Starting DMA");
  		sendChar_UART(CR);
   	sendChar_UART(LF);

	   initTimer2();
      initDMA();
      armDMA();
		timer2Start();

      while(inportw(D0TC)>0);

 	  	//timer2Stop();

  		//sleep(2);

      //outportb(0x2080, 0x0C);

      sendString_UART("Completed DMA");
  		sendChar_UART(CR);
   	sendChar_UART(LF);

		val1 = dummy[0];
      val2 = 0;
      set = 0;

      measureCount = 0;
		for (i=0; i<2000; i++)
      {
       	if ((set != 1) && (dummy[i] != val1))
         {
         	val2 = dummy[i];
         	set = 1;
         }
         if ((val1 != dummy[i]) && (val2 != dummy[i]))
         	measureCount++;
      }

		i = 0;
      pulseCount = 0;

      working = 1;
 		while (working)
   	{
      	if (i >= 2000)
         	working = 0;

   		if (dummy[i] & 0x80)
      	{
         	// Wait
         	while (dummy[i] & 0x80)
            	i++;
   		}
      	else
      	{
         	// Wait
         	while (!(dummy[i] & 0x80))
            	i++;
  	    	}

         i++;
         pulseCount++;
	   }

      for (i=0; i<2000; i++)
   			dummy[i] = 0x1;


  		sendString_UART("Distinct Measurements: ");
      measureCount += 2; // High and low + error
      sendString_UART(itoa(measureCount, buffer, 10));
	  	sendChar_UART(CR);
  	  	sendChar_UART(LF);
  		sendString_UART("Pulses Detected: ");
      sendString_UART(itoa(pulseCount, buffer, 10));
	  	sendChar_UART(CR);
  	  	sendChar_UART(LF);
  		sendString_UART("Approximate Frequency: ");
      pulseFrequency = pulseCount / sampleRate;
      sendString_UART(itoa(pulseFrequency, buffer, 10));
	  	sendChar_UART(CR);
  	  	sendChar_UART(LF);
	  	sendChar_UART(CR);
  	  	sendChar_UART(LF);


      sendString_UART("Hit any key to run another test, [Q] to quit..");
      a = getChar_UART();
      if ((a == 'Q') || (a == 'q'))
      {
 			sendChar_UART(CR);
	  		sendChar_UART(LF);
         displayMainMenu();
      }


  		sendChar_UART(CR);
   	sendChar_UART(LF);

   }
}

/*
void testUART()
{
	char a;

   while (1)
   {
		initUART();
   	sleep(1);

		// Light up red LEDs
      outportb(0x2080, 0x03);

		sendString_UART("Input a character..\0");
  		sendChar_UART(CR);
   	sendChar_UART(LF);

  		//outportb(UART_PORT1, 0x01);
      a = getChar_UART();

		sendString_UART("Character Echo.. : \0");
		sendChar_UART(a);
  		sendChar_UART(CR);
   	sendChar_UART(LF);

      sleep(1);

      // Light up green LEDs
      outportb(0x2080, 0x0C);

   	sleep(1);

   }
}
*/

void calibrateTach()
{
	char a;
   char temp;
   int pulse1k = 0;
   int pulse2k = 0;
   char buffer[10];
   int rpm1k = 0;
   int rpm2k = 0;
   int reading = 1;
   int i = 0;


 	sendChar_UART(CR);
	sendChar_UART(LF);
	sendString_UART("Tachometer Calibration");
 	sendChar_UART(CR);
	sendChar_UART(LF);
   sendString_UART("1kHz signal - Hit any key when ready, or [Q]uit:");

   a = getChar_UART();

   if ((a == 'Q') || (a == 'q'))
 	{
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
   	displayMainMenu();

      return;
	}


	sendChar_UART(CR);
	sendChar_UART(LF);

   sendString_UART("Enter calibration tach reading (1kHz): ");

   reading = 1;
   i = 0;
   while (reading)
   {
   	temp = getChar_UART();
      sendChar_UART(temp);

      if ( (temp != 'q') && (temp != 'Q') )
   	{
         buffer[i] = temp;
      	i++;
      }
      else
        	reading = 0;
 	}

   rpm1k = atoi(buffer);



	sendChar_UART(CR);
 	sendChar_UART(LF);
   sampleDMA();

   pulse1k = pulseCount;

   sendString_UART("2kHz signal - Hit any key when ready, or [Q]uit:");
   a = getChar_UART();

   if ((a == 'Q') || (a == 'q'))
 	{
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
   	displayMainMenu();

		return;
	}

 	sendChar_UART(CR);
	sendChar_UART(LF);

   sendString_UART("Enter calibration tach reading (2kHz): ");

   reading = 1;
   i = 0;
   while (reading)
   {
   	temp = getChar_UART();
      sendChar_UART(temp);

      if ( (temp != 'q') && (temp != 'Q') )
   	{
         buffer[i] = temp;
      	i++;
      }
      else
        	reading = 0;
 	}

   rpm2k = atoi(buffer);

	sendChar_UART(CR);
 	sendChar_UART(LF);
   sampleDMA();

   pulse2k = pulseCount;
   numberPulsesPerK = pulse2k - pulse1k;

	numberRPMPerK = rpm2k - rpm1k;

	sendString_UART("Pulses per count calibrated at: ");
   sendString_UART(itoa(numberPulsesPerK, buffer, 10));
 	sendChar_UART(CR);
	sendChar_UART(LF);

	sendString_UART("RPM Per K: ");
   sendString_UART(itoa(numberRPMPerK, buffer, 10));
 	sendChar_UART(CR);
	sendChar_UART(LF);

}


void displayProgramMenu()
{
 	int selectMode = 1;
	int reading = 1;
   int i = 0;
   char a;
   char temp;
   char buffer[10];

   while (selectMode)
   {
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("[R]edline RPM Level");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("[S]ampling Rate");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
      sendString_UART("[N]umber of Cylinders");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("[L]ist Current Settings");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("[C]alibrate Tachometer");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("[E]xit Programming Menu");
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
  		sendString_UART("Select Value to Program: ");
   	a = getChar_UART();
      sendChar_UART(a);
 		sendChar_UART(CR);
	  	sendChar_UART(LF);

      if ((a == 'c') || (a == 'C'))
      {
      	calibrateTach();
		}

      if ((a == 's') || (a == 'S'))
      {
 			sendChar_UART(CR);
	  		sendChar_UART(LF);

  			sendString_UART("Enter rate (5-50 ticks): ");

 			reading = 1;
         i=0;
         while (reading)
         {
          	temp = getChar_UART();
            sendChar_UART(temp);

            if ( (temp != 'q') && (temp != 'Q') )
            {
            	buffer[i] = temp;
               i++;
            }
            else
            {
            	buffer[i] = 'q';
            	reading = 0;
            }

      	}

         sampleRate = atoi(buffer);

      }

      if ((a == 'r') || (a == 'R'))
      {
 			sendChar_UART(CR);
	  		sendChar_UART(LF);
  			sendString_UART("Enter new engine redline RPM: ");

         reading = 1;
         i=0;
         while (reading)
         {
          	temp = getChar_UART();
            sendChar_UART(temp);

            if ( (temp != 'q') && (temp != 'Q') )
            {
            	buffer[i] = temp;
               i++;
            }
            else
            {
            	buffer[i] = 'q';
            	reading = 0;
				}
      	}

         redlineRPM = atoi(buffer);
      }

      if ((a == 'n') || (a == 'N'))
      {
 			sendChar_UART(CR);
	  		sendChar_UART(LF);

  			sendString_UART("Enter number of cylinders (4-8): ");

         reading = 1;
         i=0;
         while (reading)
         {
          	temp = getChar_UART();
            sendChar_UART(temp);

            if ( (temp != 'q') && (temp != 'Q') )
            {
            	buffer[i] = temp;
               i++;
            }
            else
            {
            	buffer[i] = 'q';
            	reading = 0;
				}
      	}

         numberCylinders = atoi(buffer);
      }

      if ((a == 'L') || (a == 'l'))
      {
	 		sendChar_UART(CR);
		  	sendChar_UART(LF);
  			sendString_UART("Redline RPM: ");
      	sendString_UART(itoa(redlineRPM, buffer, 10));
	  		sendChar_UART(CR);
  	  		sendChar_UART(LF);
  			sendString_UART("Number of Cylinders: ");
      	sendString_UART(itoa(numberCylinders, buffer, 10));
	  		sendChar_UART(CR);
  	  		sendChar_UART(LF);
  			sendString_UART("Sampling Rate: ");
      	sendString_UART(itoa(sampleRate, buffer, 10));
	  		sendChar_UART(CR);
  	  		sendChar_UART(LF);
  			sendString_UART("Ticks Per kHz: ");
      	sendString_UART(itoa(numberPulsesPerK, buffer, 10));
	  		sendChar_UART(CR);
  	  		sendChar_UART(LF);

      }

      if ((a == 'E') || (a == 'e'))
      {
      	selectMode = 0;
      }

   }
}

void displayActive()
{
	int working = 1;
   int numberTicks = 0;
   int i = 0;
   int j = 0;
   int temp;
   int currentRPM;
   char buffer[10];

	sendChar_UART(CR);
  	sendChar_UART(LF);

   highestRPM = 0;

   while (working)
   {
   		if (checkUserAbort())
      		working = 0;

         sampleDMA();
         numberTicks = (pulseCount / numberPulsesPerK) * 2;

         if (numberTicks > 20)
         	numberTicks = 20;

         sendString_UART("Bargraph: ");

        	for (j=0; j<numberTicks; j++ )
         {
           	sendString_UART("= ");
         }

         temp = 20 - numberTicks;

         while (temp > 0)
         {
            	sendString_UART("  ");
               temp--;
         }

         currentRPM = (pulseCount / numberPulsesPerK) * numberRPMPerK; // ((pulseCount * numberRPMPerK)  / numberPulsesPerK);

         sendString_UART(" RPM: ");
       	sendString_UART(itoa(currentRPM, buffer, 10));

         if (currentRPM > highestRPM)
         	highestRPM = currentRPM;

         if (currentRPM > redlineRPM)
         {
				// Light up red LEDs
      		outportb(0x2080, 0x03);
         }
         else
      	{
         	// Light up green LEDs
      		outportb(0x2080, 0x0C);
         }



         sendString_UART(" High RPM: ");
    		sendString_UART(itoa(highestRPM, buffer, 10));



         sendChar_UART(CR);
   }



}

void displayMainMenu()
{
	int selectMode = 1;
   char a;

   while(selectMode)
   {
  		sendString_UART("Select [P]rogram Mode, [A]ctive Mode, [D]MA Test: ");
   	a = getChar_UART();
      sendChar_UART(a);
 		sendChar_UART(CR);
	  	sendChar_UART(LF);
 /*
      if ((a == 'l') || (a == 'L'))
      {
      	displayADC();
      }
*/
      if ((a == 'P') || (a == 'p'))
      {
      	displayProgramMenu();
      }

      if ((a == 'A') || (a == 'a'))
      {
      	displayActive();
      }

      if ((a == 'D') || (a == 'd'))
      {
      	testDMA();
      }
/*
      if ((a == 'U') || (a == 'u'))
      {
      	testUART();
      }
*/
   }

}

// Initialize Perhiprial Chip Select
void initPCS()
{
   //outportb(0xffa8, 0x0007);  // MPCS
   outportw(0xffa8, 0x0087);    // MPCS
   outportw(0xffa4, 0x0207);	  // PACS
}

void main(void)
{
 	initPCS();
	initUART();

   sendChar_UART(CR);
  	sendChar_UART(LF);
   sendString_UART("Catastropic Failures Ltd. ATU ROM: 0.9.0 [BETA]\0");
	sendChar_UART(CR);
  	sendChar_UART(LF);
  	sendString_UART("Initializing System - Allowing voltages to stabilize\0");
	sendChar_UART(CR);
  	sendChar_UART(LF);
	sendChar_UART(CR);
  	sendChar_UART(LF);
   sleep(2);

	displayMainMenu();

   // Loop Forever
	while (1)
   {
   }
}
